> {- | 
>    The 'Elimination' module defines elimination orders and the
>    functions needed to perform elimination of variables. 
>  -}
> module Algebra.Elimination where 
> import qualified Data.Map as Map
> import Data.Monoid
> import Data.MonoidPlus
> import Data.MonoidRing
> import Data.Monomial
> import Data.Polynomial
> import Data.Ring
> import Algebra.GroebnerBasis

> import Debug.Trace 

> -- |We define a tuple type which inherits a product structure. The
> -- elimination order is the reverse lexicographic order on pairs. 
> newtype EliminationMonoid a b = ElimPair (a,b)
> elim_fst :: EliminationMonoid a b -> a
> elim_fst (ElimPair (a,b)) = a
> elim_snd :: EliminationMonoid a b -> b
> elim_snd (ElimPair (a,b)) = b

> elim_map :: (a -> a -> a) -> (b -> b -> b) -> EliminationMonoid a b ->
>             EliminationMonoid a b -> EliminationMonoid a b
> -- ^A helper function
> elim_map f g (ElimPair (x,y)) (ElimPair (z,t)) = 
>     ElimPair (f x z, g y t)


> instance (Monoid a, Monoid b) => 
>     Monoid (EliminationMonoid a b) where
>     mempty = ElimPair (mempty, mempty)
>     mappend = elim_map mappend mappend

> instance (Eq a, Eq b) => Eq (EliminationMonoid a b) where
>     ElimPair (a,b) == ElimPair (c,d) = (a == c) && (b == d)

Order by second variable first, so that if (a,b) <= (x,1)
then b = 1 (elimination order). 

> instance (Ord a, Ord b) => 
>     Ord (EliminationMonoid a b) where
>     compare (ElimPair (a,b)) (ElimPair (c,d)) 
>                 | test == EQ = compare a c
>                 | otherwise = test
>         where test = compare b d

> instance (OrdMonoid a, OrdMonoid b) => 
>     OrdMonoid (EliminationMonoid a b)

> instance (Show a, Show b) => 
>     Show (EliminationMonoid a b) where
>     show (ElimPair (a,b)) = shows a "*" ++ show b

> instance (GoodMonoid a, GoodMonoid b) => 
>     GoodMonoid (EliminationMonoid a b)

> instance (GroebnerMonoid a, GroebnerMonoid b) => 
>     GroebnerMonoid (EliminationMonoid a b) where 
>     (×) = mappend
>     (÷) = elim_map (÷) (÷)
>     lcm_monom = elim_map lcm_monom lcm_monom
>     lcm2_monom (ElimPair (x,y)) (ElimPair (z,t)) = 
>         (ElimPair (aplus, bplus), ElimPair (aminus, bminus))
>         where (aplus,aminus) = lcm2_monom x z
>               (bplus,bminus) = lcm2_monom y t
>     gcd_monom = elim_map gcd_monom gcd_monom
>     ElimPair (a,b) `divides` ElimPair (c,d) =
>         (a `divides` c) && (b `divides` d)

Now we define functions performing elimination. 

> separate_pols :: (Num r, VarList v) => [v] -> Polynom r v -> 
>                  MonoidRing r (EliminationMonoid (Monom v) 
>                                (Monom v))
> -- ^Convert to monomial order according to a list of variables
> separate_pols l (MakePolynom p) = q
>     where f x = x `elem` l
>           q = MakePolynom $ Map.mapKeys (ElimPair . separate_vars f) p

> eliminate :: (Fractional r, VarList v) => [v] -> Ideal (Polynom r v) ->
>              GroebnerIdeal r (Monom v)
> -- ^Given a list of variables and an ideal, eliminate these
> -- variables. 
> eliminate l (MakeIdeal i) = MakeGroebnerIdeal result
>     where elimi = map (separate_pols l) i
>           gelimi = fromGroebnerIdeal $ groebnerize $ MakeIdeal elimi

Only keep generators with eliminated variables, and reduce them to
            ordinary polynomials. 
            
>           is_pure p = (elim_snd $ leading_term p) == unit
>           result = trace ("Before elimination : " ++ show gelimi)
>                    [MakePolynom $ Map.mapKeysMonotonic (elim_fst) $
>                     fromRing p | p <- filter is_pure gelimi]

> step_eliminate :: (Fractional r, VarList v) => [v] -> Ideal (Polynom r v) ->
>              GroebnerIdeal r (Monom v)
> -- ^Eliminate variables one at a time
> step_eliminate l i | l == [] = groebnerize i
>                    | tail l == [] = eliminate l i
>                    | otherwise = step_eliminate (tail l) (MakeIdeal j)
>     where j = fromGroebnerIdeal $ eliminate [head l] i