> module Algebra.GroebnerBasis where

Basic imports.

> import Data.List
> import Data.Maybe
> import qualified Data.Map as Map

Algebra imports.

> import Data.Monoid
> import Data.MonoidPlus
> import Data.Ring
> import Data.Monomial
> import Data.MonoidRing
> import Data.Polynomial

Debugging imports

> import Debug.Trace

This module is devoted to the computation of Groebner bases. 

> newtype GroebnerIdeal r m = MakeGroebnerIdeal [MonoidRing r m]
> fromGroebnerIdeal (MakeGroebnerIdeal x) = x
> instance (Num r, GoodMonoid m) => Show (GroebnerIdeal r m) where
>     show x = show $ fromGroebnerIdeal x
  
> -- |This computes the S-polynomial of two polynomials. 
> syzygy_polynomial :: (Fractional r, GroebnerMonoid m) => 
>                      MonoidRing r m -> MonoidRing r m -> MonoidRing r m
> syzygy_polynomial f g = (MakePolynom newf) + (MakePolynom newg)
>     where (ltf,a) = leading_thing f
>           (ltg,b) = leading_thing g
>           (mf,mg) = lcm2_monom ltf ltg
>           kg = -a/b
>           newf = Map.mapKeysMonotonic (× mf) $ fromRing f
>           newg = Map.map (* kg) $ Map.mapKeysMonotonic (× mg) $ fromRing g

> -- |simple_division tries to divide the leading term. 
> simple_division :: (Fractional r, GroebnerMonoid m) => 
>                    (MonoidRing r m, m) -> (MonoidRing r m, m) -> MonoidRing r m
> simple_division (f,ltf) (g,ltg) | ltg == mempty = 0
>                                 | ltg `divides` ltf = simple_division 
>                                                       (newf,leading_term newf) (g,ltg)
>                                 | otherwise = f
>     where k = - leading_coef f / leading_coef g
>           m = ltf ÷ ltg
>           mg = toRing $ Map.map (* k) $ Map.mapKeysMonotonic (× m) $ fromRing g
>           newf = f + mg

multiv_division divides by a list

> multiv_division :: (Fractional r, GroebnerMonoid m) =>
>                    (MonoidRing r m, m) -> [(MonoidRing r m, m)] -> MonoidRing r m
> -- multiv_division (p, ltp) l | trace ("Dividing " ++ show p) False = undefined
> multiv_division (p, ltp) l | l == [] = p
>                            | p == 0  = 0
>                            | newltp == ltp = multiv_division (nextp, newltp) 
>                                              (tail l)
>                            | otherwise = nextp
>     where nextp = simple_division (p,ltp) (head l)
>           newltp = leading_term nextp

normal_form does it until the end : checking end is done by looking at leading term only.

> normal_form :: (Fractional r, GroebnerMonoid m) =>
>                (MonoidRing r m, m) -> [(MonoidRing r m, m)] -> MonoidRing r m
> normal_form (p, ltp) l | p == 0 = 0
> --                       | trace ("Dividing " ++ show p) False = undefined
>                        | ltp == nextltp = lead + normal_form (tailp, leading_term tailp) l
>                        | otherwise = normal_form (nextp, nextltp) l
>     where nextp = multiv_division (p,ltp) l
>           nextltp = leading_term nextp

Cut off f in pieces.

>           ((ld,ldt), tp) = Map.deleteFindMax $ Map.filter (/= 0) $ fromRing p
>           tailp = MakePolynom tp
>           lead = MakePolynom $ Map.singleton ld ldt 

We need to compute *reduced* syzygy polynomial

> reduced_syzygy_pol :: (Fractional r, GroebnerMonoid m) => MonoidRing r m -> 
>                    MonoidRing r m -> [(MonoidRing r m, m)] -> MonoidRing r m
> reduced_syzygy_pol a b basis 
>     = -- trace ("Syzygy for " ++ shows a " and " ++ shows b " is " ++ show result)
>       result
>     where sab = syzygy_polynomial a b
>           s = (sab, leading_term sab)
>           result = unitize (normal_form s basis)

This reduces a Groebner basis : first we check for polynomials with
     dividing leading terms (wrong_n) then reduce them one w.r.t
     others.

> self_normalize l = reducedl
>     where reducedl = map unitize (filter (/= 0) [normal_form f (l \\ [f]) | f <- l])

An iteration of self_normalize.

> full_normalize l | newlt == map snd l = newp
>                  | otherwise          = full_normalize newl
>     where newp = self_normalize l
>           newl = map (\s -> (s, leading_term s)) newp
>           newlt = map snd newl

> reduce_groebner_basis :: (Fractional r, GroebnerMonoid m) => 
>                          GroebnerIdeal r m -> GroebnerIdeal r m
> reduce_groebner_basis (MakeGroebnerIdeal l) 
>                           | l == [] = MakeGroebnerIdeal []
> --                          | trace ("Reduced length " ++ shows 
> --                                   (length l) " basis to length "
> --                                   ++ shows (length reduced) ".") False = undefined
>                           | otherwise = MakeGroebnerIdeal (sortBy compareByLT reduced)
>     where ltl = map (\s -> (s, leading_term s)) (nub (filter (/= 0) l))
>           test (a,b) = (snd a) `divides` (snd b)
>           wrong = map snd $ filter test [(x,y) | x <- ltl, y <- ltl \\ [x]]
>           newl = ltl \\ wrong 
>           reduced = self_normalize newl
>           compareByLT a b = compare (leading_term a) (leading_term b)

This takes a partial Groebner basis and a list of missing syzygies,
     and eliminate non-minimal leading terms (for divisibility order) 
     and the corresponding syzygies. 

     cleanup_nonminimal basis syzs = (newbasis, newsyzs)

     For any LT(f) divides LT(g), g can be replaced by S(f,g), and
     (g,h) syzygies should be replaced by (S,h) syzygies. We only test for 
     f = head of basis.

     Output syzygies ordered by difficulty

> cleanup_nonminimal :: (Fractional r, GroebnerMonoid m) => 
>                       [(MonoidRing r m, m)] -> [(MonoidRing r m, MonoidRing r m)] 
>                    -> ([(MonoidRing r m, m)], [(MonoidRing r m, MonoidRing r m)])
> cleanup_nonminimal basis syzs       
>     | wrong == Nothing = -- trace "Cleaned up. Current monomials " 
>                          -- ++ show (sort $ map snd basis))
>                          (basis, sortBy compare_difficulty syzs)
>     | otherwise        = cleanup_nonminimal newbasis newsyzs
>     where
>     purebasis = basis -- filter (\x@(t,lt) -> (t /= 0)) basis
>     test (a,b) = ((snd a) `divides` (snd b)) -- && (fst a /= 0)
>     wrong = find test [(f,g) | f <- basis, g <- basis \\ [f]]

The following will only be calculated if wrong is non-empty: sfg is
the syzygy polynomial, badsyzs[12] are syzgies having g as first/second term. 

>     realwrong@(wf,wg) = fromJust wrong
>     f = fst wf
>     g = fst wg
>     sfg = trace (shows (snd wf) " divides " 
>                                ++ shows (snd wg) ".")
>           reduced_syzygy_pol f g purebasis
>     newbasis | sfg == 0  = purebasis \\ [wg]
>              | otherwise = (sfg,leading_term sfg):(purebasis \\ [wg])
>     goodsyzs = filter (\x -> ((fst x /= g) && (snd x /= g))) syzs
>     newsyzs | sfg == 0 = goodsyzs
>             | otherwise = goodsyzs ++ [(sfg, fst h) | h <- (tail newbasis)]

> difficulty :: (Num r, GroebnerMonoid m) => MonoidRing r m -> MonoidRing r m -> m
> difficulty p q = lcm_monom (leading_term p) (leading_term q)

> compare_difficulty (a,b) (c,d) = compare (difficulty a b) (difficulty c d)

> {- |
>    The process_syzygies function takes a list of generators and a list of
>    pairs whose syzygies should be added to the Gröbner basis. Syzygies
>    are supposed to come ordered by diffficulty.  
>  -}
> process_syzygies :: (Fractional r, GroebnerMonoid m) => [(MonoidRing r m, m)] ->
>                      [(MonoidRing r m, MonoidRing r m)] -> [(MonoidRing r m, m)]
> process_syzygies current l | l == []   = current
>                            | (length syzs) `mod` 10 == 0 && trace 
>                              (shows (length syzs) " syzygies left.\r") False = undefined
>                            | s == 0    = process_syzygies current syzs
>                            | otherwise = (uncurry process_syzygies) 
>                                          (cleanup_nonminimal newbasis newsyzs)
>     where t@(p,q) = head l 
>           syzs = tail l
>           ltp = leading_term p; ltq = leading_term q

Case where we do not compute s : ltp and ltq are coprime. 
Other case : lcm(ltp, ltq) is a multiple of ltk where (p,k), (q,k) are no remaining syzs. 

>           lcmlt = lcm_monom ltp ltq
>           divisors = map fst $ filter (\k -> snd k `divides` lcmlt) current
>           excellentdivisor = find (\k -> (intersect syzs 
>                                           [(k,p),(k,q),(p,k),(q,k)]) == []) divisors
>           s | gcd_monom ltp ltq == mempty = 0
>           -- trace ("Skipping syz : coprime lts") True = 0
>             | excellentdivisor == Nothing = 0
>           --    trace ("Skipping syz : found " ++ show excellentdivisor) 
>             | otherwise = reduced_syzygy_pol p q current
>           lts = leading_term s

The following only if s /= 0. Newsyzs is sorted in cleanup function. 

>           newsyzs = trace ("Adding element to basis. " ++ shows 
>                            (length current) " already there.") 
>                     syzs ++ [(s,fst c) | c <- current]
>           newbasis = (s,lts):current 

grobnerize put initial polynomials, chooses syzygies and calls process_syzygies. 

> groebnerize :: (Fractional r, GroebnerMonoid m) => Ideal (MonoidRing r m) -> GroebnerIdeal r m
> groebnerize (MakeIdeal l) | trace ("Groebnerizing " ++ show l) False = undefined
>                           | otherwise = reduce_groebner_basis $ MakeGroebnerIdeal $ map fst raw_groebner
>     where purel = nub (filter (/= 0) l) 
>           maxidx = (length purel) - 1
>           ltl = map (\x -> (x, leading_term x)) purel
>           raw_groebner = process_syzygies ltl 
>                          [(purel!!x,purel!!y) | x <- [0..maxidx], y <- [0..x-1]]

> check_groebnerness :: (Fractional r, GroebnerMonoid m) => 
>                       GroebnerIdeal r m -> Maybe (GroebnerIdeal r m)
> check_groebnerness (MakeGroebnerIdeal l) 
>                        | bad == Nothing = Just (MakeGroebnerIdeal l)
>                        | otherwise = trace ("Bad syzygy : " 
>                                             ++ show (fromJust bad)) 
>                                      Nothing
>     where ltl = [(f, leading_term f) | f <- l]
>           syzs [] = []
>           syzs l = [(f,g) | f <- [head l], g <- tail l] ++ syzs (tail l)
>           p = [reduced_syzygy_pol f g ltl | s@(f,g) <- syzs l]
>           bad = find (/= 0) p

