> module Data.MonoidRing where
> import Data.List
> import qualified Data.Map as Map
> import Data.Monoid
> import Data.MonoidPlus
> import Data.Ring

Given a datatype of class Num (a coefficient ring) and an ordered
      monoid, we represent the bifunctor MonoidRing using binary tree
      maps m -> a(m) tracking coefficients of monomials.

> data (Num r, OrdMonoid m) => MonoidRing r m = MakePolynom (Map.Map m r)
> toRing x = MakePolynom x
> fromRing (MakePolynom x) = x

The algebra of a monoid is equipped with addition (ring_sum) and
    product (ringprod, defined inductively starting with monomials). 

> zero_ring :: (Num r, OrdMonoid m) => MonoidRing r m 
> zero_ring = MakePolynom (Map.empty)

> ringsum :: (Num r, OrdMonoid m) => 
>            MonoidRing r m -> MonoidRing r m -> MonoidRing r m 
> ringsum (MakePolynom x) (MakePolynom y) 
>     = MakePolynom (Map.filter (/= 0) (Map.unionWith (+) x y))

> ringprod :: (Num r, OrdMonoid m) => 
>             MonoidRing r m -> MonoidRing r m -> MonoidRing r m 
> ringprod (MakePolynom x) (MakePolynom y) = 
>     MakePolynom (Map.filter (/= 0) (Map.fromListWith (+) l))
>     where a = Map.toAscList x
>           b = Map.toAscList y
>           l = [(mappend m n, r * s) | (m, r) <- a, (n, s) <- b]

A monad-like structure is given by specifying how choosing images of
  generators defines a morphism of algebras (should we check that it
  is a morphism of monoids ?). 

> smallmap :: (Num r, OrdMonoid m, OrdMonoid n) =>
>             (r,m) -> (m -> MonoidRing r n) -> MonoidRing r n
> smallmap (r,m) f = MakePolynom (Map.map (* r) fm) where
>     fm = fromRing (f m)

> (>>°) :: (Num r, OrdMonoid m, OrdMonoid n) =>
>      (MonoidRing r m) -> (m -> MonoidRing r n) -> MonoidRing r n 
> -- ^Monadic map of rings R[M] -> R[N] specified by a morphism M -> R[N]
> (MakePolynom x) >>° f = Map.foldWithKey mapf zero_ring x where
>     mapf m a y = (smallmap (a,m) f) `ringsum` y

> returnr :: (Num r, OrdMonoid m) => m -> MonoidRing r m
> returnr x = MakePolynom (Map.singleton x 1)

The total order on the monoid defines equality.

> instance (Num r, OrdMonoid m) => Eq (MonoidRing r m) where
>     MakePolynom x == MakePolynom y = 
>         Map.filter (/= 0) x == Map.filter (/= 0) y

> instance (Num r, Ord r, OrdMonoid m) => Ord (MonoidRing r m) where
>     compare (MakePolynom a) (MakePolynom b) = compare a b

MonoidRing are instances of the Ring class.

> instance (Num r, OrdMonoid m) => Ring (MonoidRing r m) where
>     (+°) = ringsum
>     (*°) = ringprod
>     minus (MakePolynom x) = MakePolynom (Map.map negate x)
>     fromInt n = MakePolynom (Map.singleton mempty (fromInteger n))

Ring product turns monoid algebras into monoids.

> instance (Num r, OrdMonoid m) => Monoid (MonoidRing r m) where
>     mempty = returnr mempty
>     mappend = ringprod

> instance (Num r, Ord r, OrdMonoid m) => OrdMonoid (MonoidRing r m) where

> leading_thing :: (Num r, OrdMonoid m) => MonoidRing r m -> (m, r)
> leading_thing p | fp == Map.empty = (mempty, 0)
>                 | otherwise       = result
>     where fp = fromRing p
>           mp = Map.findMax fp
>           result | snd mp == 0 = leading_thing (MakePolynom $ Map.deleteMax fp)
>                  | otherwise = mp

> -- |The leading_term function selects the leading monomial, taking the
> -- maximal leaf of the associated binary tree.
> leading_term :: (Num r, OrdMonoid m) => MonoidRing r m -> m
> leading_term = fst . leading_thing

> leading_coef :: (Num r, OrdMonoid m) => MonoidRing r m -> r
> leading_coef = snd . leading_thing

> unitize :: (Fractional r, OrdMonoid m) => MonoidRing r m -> MonoidRing r m
> unitize f | f == zero_ring = zero_ring
>           | otherwise = MakePolynom (Map.map (/ a) (fromRing f)) where a = leading_coef f

Algebras of graded monoids are themselves graded rings. 

> instance (Num r, OrdMonoid m, GradedMonoid m) 
>     => GradedMonoid (MonoidRing r m) where 
>     degree p = degree (leading_term p)

> degree_pol :: (Num r, OrdMonoid m, GradedMonoid m) 
>               => MonoidRing r m -> Integer
> -- ^The 'degree_pol' function returns the degree of an element. 
> degree_pol p = degree p

This altogether with the Show structure (see later) allows to define a
     Num instance structure on monoid algebras. 

> instance (Num r, GoodMonoid m) => Num (MonoidRing r m) where
>     fromInteger n = fromInt n
>     (+) = ringsum
>     (*) = ringprod
>     negate = minus
>     abs (MakePolynom x) = MakePolynom (Map.map abs x)
>     signum (MakePolynom x) = MakePolynom (Map.map signum x)

Given a Show structure of the base monoid, define a Show structure on
      its algebra using additive notation. Output will look like 
      "3*X^2*Y+4*X*Z". 

> frobnicate_add (m,r) s | r == 1    = shows m s 
>                        | r == -1   = '-':shows m s 
>                        | otherwise = shows r ('*':shows m s)
> prettyshow_add :: (Num r, Show m, Ord m) => [(m, r)] -> String
> prettyshow_add []   = "0"
> prettyshow_add (l:ls) | ls == []  = frobnicate_add l ""
>                       | otherwise = frobnicate_add l ('+':prettyshow_add ls)

> instance (Num r, GoodMonoid m) => Show (MonoidRing r m) where 
>     show (MakePolynom x) = prettyshow_add (reverse l) where
>         l = Map.toAscList (Map.filter (/= 0) x)

