> {- |
>    The 'Monomial' module defines the basic classes and structures
>    for computing with free commutative monoids over a finite totally
>    ordered set. 
>  -}
> module Data.Monomial where
> import Data.List
> import Data.Maybe
> import Data.Array.IArray
> import Data.Array.Unboxed
> import Data.Monoid
> import Data.MonoidPlus
> import Data.Lattice

> import Debug.Trace 

> -- |The 'VarList' class inherits structure from suitable classes of
> -- totally ordered, finite types, with Show structure
> class (Ix a, Bounded a, Ord a, Enum a, Show a) => VarList a where
>     varBounds :: (a,a)
>     varBounds = (minBound, maxBound)
>     varRange :: [a]
>     varRange = range varBounds

> {- |
>    The Monomial module defines free abelian semi-groups over a
>    countable type: since arrays will be used, the Ix class is used
>    for the variables. We use unboxed array types.
>    
>    It includes a field holding the degree. 
>  -}

> data (VarList a) => Monom a = MakeMonom (Array a Int, Integer)
> fromMonom (MakeMonom l) = fst l
> toMonom :: (VarList a) => Array a Int -> Monom a
> toMonom l = MakeMonom (l, toInteger $ foldr (+) 0 $ elems l)

The free abelian semi-group has a unit, corresponding to the empty
    list, base variables corresponding to elements of T. Its product
    is based on map fusion, combinate duplicate variables with + 
    (adding exponents).

> -- |'unit' is the unit monomial, which has zero exponents
> unit :: (VarList a) => Monom a
> unit = MakeMonom (array varBounds [(i,0) | i <- varRange], 0)
> -- |'returnm' is the monad-like unit of the functor 'Monom'
> returnm :: (VarList a) => a -> Monom a
> returnm x = MakeMonom (fromMonom unit // [(x,1)], 1)

Monomials are instances of GroebnerMonoids, with product, quotient,
          and lcm. 

> instance (VarList a) => GroebnerMonoid (Monom a) where 
>     MakeMonom (x,dx) × MakeMonom (y,dy) = MakeMonom l
>         where l = (accum (+) x (assocs y), dx+dy) 
>     MakeMonom (x,dx) ÷ MakeMonom (y,dy) = MakeMonom l
>         where l = (accum (-) x (assocs y), dx-dy) 
>     lcm_monom x y = toMonom l
>         where l = accum (max) (fromMonom x) (assocs (fromMonom y))
>     lcm2_monom x y = (d ÷ x, d ÷ y)
>         where d = lcm_monom x y 
>     gcd_monom x y = toMonom l
>         where l = accum (min) (fromMonom x) (assocs (fromMonom y)) 
>     x `divides` y = and $ map (uncurry (<=)) l
>         where l = zip (elems (fromMonom x)) (elems (fromMonom y))

If x is divisible by y, it is easy to divide them : 

> unsafe_division :: (VarList a) => Monom a -> Monom a -> Monom a
> unsafe_division = (÷)

We define helper functions : raising to a power (××), map a monomial
   x^n in N[A] into another semi-group N[B] using a function A -> N[B]
   (smallmap function), and the fullmap N[A] -> N[B] defined by the
   images of generators. 

> -- |Helper function '(××)' raising to an integral power. 
> (××) :: (VarList a) => Monom a -> Int -> Monom a
> x ×× n = MakeMonom (amap (* n) (fromMonom x), (toInteger n) * (degree x))

> smallmap :: (VarList b) => (a,Int) -> (a -> Monom b) -> Monom b
> -- ^Monadic partial operation.
> smallmap (x,n) f = (f x) ×× n

> -- |A monadic operation computing semi-group maps from an assignment
> -- of images of generators.  
> (>>!) :: (VarList a, VarList b) => Monom a -> 
>          (a -> Monom b) -> Monom b
> x >>! f = foldr mapf unit $ assocs $ fromMonom x where 
>    mapf xn y = (smallmap xn f) × y

The functor Monom is actually a Monad, which should have the following
    properties:

instance Monad Monom where
     (>>=) = fullmap
     return = returnm

Free semigroups are instances of the classes Monoid and custom
     OrdMonoid. 

> instance (VarList a) => Eq (Monom a) where
>     x == y = (fromMonom x) == (fromMonom y)

> instance (VarList a) => Monoid (Monom a) where
>     mempty = unit
>     mappend = (×)

> instance (VarList a) => OrdMonoid (Monom a)

> instance (VarList a) => GradedMonoid (Monom a) where
>     degree (MakeMonom l) = snd l 

> {- |
>    Using ordered variables allows to define equality and lexicographic
>    ordering using predefined instances for maps. The corresponding
>    order is total degree inverse lexicographic order : x is greater
>    than y if there are more last variables in y.
>  -}
> -- "c" is the last differing exponent.
> instance (VarList a) => Ord (Monom a) where 
>     compare x y | raw /= EQ = raw
>                 | c /= Nothing = compare ly lx
>                 | otherwise    = EQ
>         where raw = compare (degree x) (degree y)
>               expx = elems $ fromMonom x
>               expy = elems $ fromMonom y
>               c = find (uncurry (/=)) $ reverse $ zip expx expy
>               (lx,ly) = fromJust c

When variables have a Show method, the free semi-group inherits a Show
     method taking [(X,2),(Y,1)] to "X²*Y". The frobnicate function
     takes care of elementary monomials x^n, while prettyshow is the
     main block. 

 exponent_char :: Char -> String
 exponent_char c | c=='0' = "⁰"
                 | c=='1' = "¹"
                 | c=='2' = "²"
                 | c=='3' = "³"
                 | c=='4' = "⁴"
                 | c=='5' = "⁵"
                 | c=='6' = "⁶"
                 | c=='7' = "⁷"
                 | c=='8' = "⁸"
                 | c=='9' = "⁹"
                 | otherwise = [c]

> exponent_char s = '^':s

> frobnicate (x,n) s | n == 1    = shows x s
>                    | otherwise = shows x ((exponent_char $ show n) ++ s)
> prettyshow :: (Show a, VarList a) => [(a, Int)] -> String
> prettyshow [] = "1"
> prettyshow (l:ls) | ls == []  = frobnicate l ""
>                   | otherwise = frobnicate l ('*':prettyshow ls)

This allows to define a Show structure and a GoodMonoid structure on
     free semi-groups over (Ord, Show). 

> instance (Show a, VarList a) => Show (Monom a) where
>     show m = prettyshow $ filter (\x -> snd x /= 0) $ assocs $ fromMonom m
> instance (Show a, VarList a) => GoodMonoid (Monom a)

> separate_vars :: (VarList v) => (v -> Bool) -> Monom v ->
>                  (Monom v, Monom v)
> -- ^A helper function for elimination, taking a boolean function
> -- indicating which variables should be eliminated. 
> separate_vars f x = (toMonom y, toMonom z)
>     where right s@(t,x) = if f t then (t,x) else (t,0)
>           left s@(t,x) = if f t then (t,0) else (t,x)
>           y = array varBounds $ map left $ assocs $ fromMonom x
>           z = array varBounds $ map right $ assocs $ fromMonom x